/*!
 * Copyright (C) 2006 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

/*
 * Limited AT91SAM9260-EK MP3 player sample.
 *
 * Will sequentially play all MP3 files from an MMC's root directory.
 *
 * MultiMedia Cards only, no SD Cards (see arch/arm/dev/at91_mmc.c).
 * Root directory must contain MP3 files only.
 * Hardcoded for 180MHz CPU clock and 44.1kHz sampling rate (also see ./at73dac.c).
 */

#include <cfg/os.h>
#include <cfg/clock.h>
#include <dev/board.h>
#include <sys/heap.h>
#include <sys/timer.h>
#include <sys/event.h>

#include <stdlib.h>
#include <string.h>

#include <stdio.h>

#include "at73dac.h"
#include "receiver.h"
#include "mp3player.h"

/*!
 *
 */
THREAD(Mp3PlayerThread, arg)
{
    PLAYERINFO *pip = (PLAYERINFO *)arg;
    RECEIVERINFO *rip;
    int rbytes;
    u_char *bufptr;
    int ec;
    int errcnt;

    for (;;) {
        /* Idle loop. */
        for (;;) {
            printf("[PIDLE]");
            pip->pi_status = PSTAT_IDLE;
            NutEventBroadcast(&pip->pi_stsevt);

            /* Wait for start event. */
            NutEventWait(&pip->pi_cmdevt, 0);
            printf("[PEVT%u]", pip->pi_status);
            if (pip->pi_status == PSTAT_START) {
                if (pip->pi_mp3dec == NULL) {
                    pip->pi_mp3dec = MP3InitDecoder();
                }
                if (pip->pi_mp3dec && pip->pi_rcvr) {
                    break;
                }
            }
        }

        /* Broadcast running event. */
        printf("[PRUN]");
        pip->pi_status = PSTAT_RUNNING;
        NutEventBroadcast(&pip->pi_stsevt);
        rip = pip->pi_rcvr;

        /* Wait for enough data to start playing. */
        NutSleep(500);
        while (rip->ri_avail < 2 * MAINBUF_SIZE) {
            if (NutEventWait(&rip->ri_rdbque, 5000)) {
                /* Timeout. Go back to idle state (see below). */
                break;
            }
        }

        /* Running loop. */
        errcnt = 0;
        while (pip->pi_status == PSTAT_RUNNING) {
            /* Wait for data for a short period only. */
            if (rip->ri_avail < 2 * MAINBUF_SIZE) {
                if (NutEventWait(&rip->ri_rdbque, 500)) {
                    break;
                }
                continue;
            }

            if (rip->ri_rpos > rip->ri_wpos && MP3_BUFSIZ - rip->ri_rpos < 2 * MAINBUF_SIZE) {
                /* The read pointer is in front of the write pointer, but
                   there is not enough MP3 data at the end of the buffer.
                   Wrap data from the front to the back, because the decoder
                   requires a continous buffer. Note, that we allocated the
                   additional bytes at the end of the MP3 data buffer for
                   this purpose. */
                memcpy(&rip->ri_buff[MP3_BUFSIZ], rip->ri_buff, 2 * MAINBUF_SIZE - (MP3_BUFSIZ - rip->ri_rpos));
            }

            /* Test for sync error. Discard data up to the next sync word. */
            if ((rbytes = MP3FindSyncWord(&rip->ri_buff[rip->ri_rpos], 2 * MAINBUF_SIZE)) < 0) {
                printf("[ERR-SYN]");
			    break;
		    }
            else if (rbytes == 0) {
                /* We are in sync. Decode the next MP3 frame. */
                rbytes = 2 * MAINBUF_SIZE;
                bufptr = &rip->ri_buff[rip->ri_rpos];
 		        ec = MP3Decode(pip->pi_mp3dec, &bufptr, &rbytes, pip->pi_pcmbuf, 0);
                rbytes = 2 * MAINBUF_SIZE - rbytes;
                if (ec == 0) {
                    /* Successful decoding. Retrieve the frame info to determine the
                       the number of output samples. */
                    MP3GetLastFrameInfo(pip->pi_mp3dec, &pip->pi_frameinfo);
                    if (At73DacWrite(pip->pi_pcmbuf, pip->pi_frameinfo.outputSamps)) {
                        printf("[ERR-ADC]");
                        break;
                    }
                    /* Slowly recover from previous errors. */
                    if (errcnt) {
                        errcnt--;
                    }
                }
                else {
                    printf("\n[ERR%d]%lu bytes free", ec, NutHeapAvailable());
                    /* Move to idle state in case of fatal errors. */
                    if (ec == ERR_MP3_OUT_OF_MEMORY) {
                        break;
                    }
                    /* In case of too many errors move to idle state. */
                    if (++errcnt > 8) {
                        break;
                    }
                    /* Avoid to get stuck on the same error. */
                    if (rbytes == 0) {
                        rbytes = 1;
                    }
                }
            }
            else {
                printf("\n[S%d,r=%u,w=%u ", rbytes, rip->ri_rpos, rip->ri_wpos);
                printf("%02X %02X %02X]", rip->ri_buff[rip->ri_rpos - 1], 
                    rip->ri_buff[rip->ri_rpos], rip->ri_buff[rip->ri_rpos + 1]);
                /* In case of too many sync errors move to idle state. */
                if (++errcnt > 8) {
                    break;
                }
            }
            /* Adjust the read pointer and the number of bytes available. */
            rip->ri_avail -= rbytes;
            rip->ri_rpos += rbytes;
            if (rip->ri_rpos >= MP3_BUFSIZ) {
                rip->ri_rpos -= MP3_BUFSIZ;
            }
            /* Inform the receiver, that we removed data from the buffer. */
            NutEventPost(&rip->ri_wrbque);
        }
    }
}

int Mp3PlayerInit(PLAYERINFO *pip)
{
    if (pip->pi_tid == 0) {
        pip->pi_tid = NutThreadCreate("mp3play", Mp3PlayerThread, pip, 4096);
    }
    return pip->pi_tid ? 0 : -1;
}
